@@ -27,10 +27,13 @@
 import os
 import shutil
 import subprocess
+import sys
 import tempfile
 
-from PyQt5.QtCore import QDir, QFile, QFileInfo, QIODevice
+from PyQt5.QtCore import QDir, QFile
 
+from ..file_utilities import (create_file, get_embedded_dir,
+        get_embedded_file_for_version, read_embedded_file)
 from ..metadata import pyqt4_metadata, pyqt5_metadata
 from ..project import QrcDirectory, QrcFile
 from ..user_exception import UserException
@@ -40,17 +43,15 @@ from ..version import PYQTDEPLOY_HEXVERSION
 class Builder():
     """ The builder for a project. """
 
-    def __init__(self, project):
+    def __init__(self, project, message_handler):
         """ Initialise the builder for a project. """
 
         super().__init__()
 
         self._project = project
+        self._message_handler = message_handler
 
-        self.quiet = False
-        self.verbose = False
-
-    def build(self, build_dir=None, clean=False, console=False):
+    def build(self, opt, build_dir=None, clean=False, console=False):
         """ Build the project in a given directory.  Raise a UserException if
         there is an error.
         """
@@ -65,14 +66,15 @@ class Builder():
             build_dir = project.absolute_path(build_dir)
 
         if clean:
-            self._progress("Cleaning {0}.".format(build_dir))
+            self._message_handler.progress_message(
+                    "Cleaning {0}".format(build_dir))
             shutil.rmtree(build_dir, ignore_errors=True)
 
         self._create_directory(build_dir)
 
         freeze = self._copy_lib_file('freeze.py')
 
-        self._write_qmake(build_dir, console, freeze)
+        self._write_qmake(build_dir, console, freeze, opt)
 
         resources_dir = os.path.join(build_dir, 'resources')
         self._create_directory(resources_dir)
@@ -81,16 +83,21 @@ class Builder():
         for resource in self.resources():
             if resource == '':
                 package = project.application_package
-                self._write_resource(resources_dir, resource, package, 
-                        project.absolute_path(package.name), freeze)
+                contents = self._write_package(resources_dir, resource,
+                        package, project.absolute_path(package.name), freeze,
+                        opt)
+                self._write_resource_qrc(resources_dir, 'pyqtdeploy.qrc',
+                        contents)
             elif resource == 'stdlib':
-                self._write_resource(resources_dir, resource,
+                contents = self._write_package(resources_dir, resource,
                         project.stdlib_package, os.path.join(stdlib_dir, ''),
-                        freeze)
+                        freeze, opt)
+                self._write_resource_qrc(resources_dir, 'stdlib.qrc',
+                        contents)
             else:
                 # Add the PyQt package to a temporary copy of the site-packages
                 # package.
-                site_packages_package = project.site_packages_package.copy()
+                site_packages_package = project.packages[0].copy()
                 pyqt_dir = 'PyQt5' if project.application_is_pyqt5 else 'PyQt4'
 
                 if len(project.pyqt_modules) != 0:
@@ -106,19 +113,30 @@ class Builder():
                                         pyqt_dir),
                                 'uic', [])
 
-                self._write_resource(resources_dir, resource,
+                contents = self._write_package(resources_dir, resource,
                         site_packages_package,
-                        os.path.join(stdlib_dir, 'site-packages', ''), freeze)
+                        os.path.join(stdlib_dir, 'site-packages', ''), freeze,
+                        opt)
+
+                for package in project.packages[1:]:
+                    contents += self._write_package(resources_dir, resource,
+                            package,
+                            os.path.join(project.absolute_path(package.name),
+                                    ''),
+                            freeze, opt)
+
+                self._write_resource_qrc(resources_dir, 'site-packages.qrc',
+                        contents)
 
         os.remove(freeze)
 
-    def _add_uic_dir(self, package, pyqt_dir, dirname, dir_stack):
+    def _add_uic_dir(self, package, pyqt_dir, dir_name, dir_stack):
         """ Add a uic directory to a package. """
 
-        dir_pkg = QrcDirectory(dirname)
+        dir_pkg = QrcDirectory(dir_name)
         package.contents.append(dir_pkg)
 
-        dirpath = [pyqt_dir] + dir_stack + [dirname]
+        dirpath = [pyqt_dir] + dir_stack + [dir_name]
         dirpath = os.path.join(*dirpath)
 
         for content in os.listdir(dirpath):
@@ -130,58 +148,84 @@ class Builder():
             if os.path.isfile(content_path):
                 dir_pkg.contents.append(QrcFile(content))
             elif os.path.isdir(content_path):
-                dir_stack.append(dirname)
+                dir_stack.append(dir_name)
                 self._add_uic_dir(dir_pkg, pyqt_dir, content, dir_stack)
                 dir_stack.pop()
 
-    def _write_qmake(self, build_dir, console, freeze):
+    def _write_qmake(self, build_dir, console, freeze, opt):
         """ Create the .pro file for qmake. """
 
         project = self._project
 
-        app_name = project.application_basename()
-
-        f = self._create_file(build_dir, app_name + '.pro')
+        f = self._create_file(build_dir,
+                project.application_basename() + '.pro')
 
         f.write('TEMPLATE = app\n')
-        f.write('CONFIG += release warn_on\n')
-
-        # Configure the QT value.
         f.write('\n')
 
-        no_gui = True
-        qmake_qt = []
-        qmake_config = []
+        # Configure the CONFIG and QT values.
+        needs_gui = False
+        qmake_qt4 = set()
+        qmake_config4 = set()
+        qmake_qt5 = set()
+        qmake_config5 = set()
 
         for pyqt_m in project.pyqt_modules:
-            needs_gui = True
-
             metadata = self._get_module_metadata(pyqt_m)
 
-            for qt in metadata.qt:
-                if qt == '-gui':
-                    needs_gui = False
-                elif qt not in qmake_qt:
-                    qmake_qt.append(qt)
+            if metadata.gui:
+                needs_gui = True
 
-            for config in metadata.config:
-                if config not in qmake_config:
-                    qmake_config.append(config)
+            qmake_qt4.update(metadata.qt4)
+            qmake_config4.update(metadata.config4)
+            qmake_qt5.update(metadata.qt5)
+            qmake_config5.update(metadata.config5)
 
-            if needs_gui:
-                no_gui = False
+        both_qt = qmake_qt4 & qmake_qt5
+        qmake_qt4 -= both_qt
+        qmake_qt5 -= both_qt
 
-        if no_gui or console:
-            f.write('CONFIG += console\n')
+        both_config = qmake_qt4 & qmake_qt5
+        qmake_config4 -= both_config
+        qmake_config5 -= both_config
 
-        if no_gui:
+        both_config.add('release')
+        both_config.add('warn_on')
+
+        if not needs_gui or console:
+            both_config.add('console')
+
+        f.write('CONFIG += {0}\n'.format(' '.join(both_config)))
+
+        if not needs_gui:
             f.write('QT -= gui\n')
 
-        if len(qmake_qt) != 0:
-            f.write('QT += {0}\n'.format(' '.join(qmake_qt)))
+        if both_qt:
+            f.write('QT += {0}\n'.format(' '.join(both_qt)))
 
-        if len(qmake_config) != 0:
-            f.write('CONFIG += {0}\n'.format(' '.join(qmake_config)))
+        if qmake_config4 or qmake_qt4:
+            f.write('\n')
+            f.write('lessThan(QT_MAJOR_VERSION, 5) {\n')
+
+            if qmake_config4:
+                f.write('    CONFIG += {0}\n'.format(' '.join(qmake_config4)))
+
+            if qmake_qt4:
+                f.write('    QT += {0}\n'.format(' '.join(qmake_qt4)))
+
+            f.write('}\n')
+
+        if qmake_config5 or qmake_qt5:
+            f.write('\n')
+            f.write('greaterThan(QT_MAJOR_VERSION, 4) {\n')
+
+            if qmake_config5:
+                f.write('    CONFIG += {0}\n'.format(' '.join(qmake_config5)))
+
+            if qmake_qt5:
+                f.write('    QT += {0}\n'.format(' '.join(qmake_qt5)))
+
+            f.write('}\n')
 
         # Determine the extension modules and link against them.
         extensions = {}
@@ -253,15 +297,8 @@ class Builder():
             f.write('LIBS += -L{0} -l{1}\n'.format(self._quote(lib_dir), lib))
 
         # Add the platform specific stuff.
-        platforms_f = QFile(self._lib_filename('platforms.pro'))
-
-        if not platforms_f.open(QIODevice.ReadOnly|QIODevice.Text):
-            self._error(
-                    "Unable to open file {0}.".format(platforms_f.fileName()),
-                    platforms_f.errorString())
-
-        platforms = platforms_f.readAll()
-        platforms_f.close()
+        platforms = read_embedded_file(
+                self._get_lib_file_name('platforms.pro'))
 
         f.write('\n')
         f.write(platforms.data().decode('latin1'))
@@ -285,25 +322,29 @@ class Builder():
         # Specify the source and header files.
         f.write('\n')
 
-        f.write('SOURCES = main.c pyqtdeploy_main.c pyqtdeploy_module.cpp\n')
-        self._write_main_c(build_dir, app_name, extensions.keys())
-        self._copy_lib_file('pyqtdeploy_main.c', build_dir)
+        f.write('SOURCES = main.c pyqtdeploy_start.cpp pyqtdeploy_module.cpp\n')
+        self._write_main_c(build_dir, extensions.keys())
+        self._copy_lib_file('pyqtdeploy_start.cpp', build_dir)
         self._copy_lib_file('pyqtdeploy_module.cpp', build_dir)
 
         f.write('HEADERS = frozen_bootstrap.h frozen_main.h pyqtdeploy_version.h\n')
 
-        major, _ = project.python_target_version
+        major, minor = project.python_target_version
         if major is None:
-            self._error("Unable to determine target Python version")
+            raise UserException("Unable to determine target Python version")
 
-        bootstrap = self._copy_lib_file(
-                'bootstrap_py3.py' if major == 3 else 'bootstrap_py2.py')
+        py_version = (major << 16) + (minor << 8)
+
+        bootstrap_src = get_embedded_file_for_version(py_version, __file__,
+                'lib', 'bootstrap')
+        bootstrap = self._copy_lib_file(bootstrap_src,
+                dst_file_name='bootstrap.py')
         self._freeze(os.path.join(build_dir, 'frozen_bootstrap.h'), bootstrap,
-                freeze, name='pyqtdeploy_bootstrap')
+                freeze, opt, name='pyqtdeploy_bootstrap')
         os.remove(bootstrap)
 
         self._freeze(os.path.join(build_dir, 'frozen_main.h'),
-                project.absolute_path(project.application_script), freeze,
+                project.absolute_path(project.application_script), freeze, opt,
                 name='pyqtdeploy_main')
 
         version_f = self._create_file(build_dir, 'pyqtdeploy_version.h')
@@ -329,13 +370,19 @@ class Builder():
 
         resources.append('stdlib')
 
+        need_site_packages = False
+
         if len(project.pyqt_modules) != 0:
-            resources.append('site-packages')
+            need_site_packages = True
         else:
-            for content in project.site_packages_package.contents:
-                if content.included:
-                    resources.append('site-packages')
-                    break
+            for package in project.packages:
+                for content in package.contents:
+                    if content.included:
+                        need_site_packages = True
+                        break
+
+        if need_site_packages:
+            resources.append('site-packages')
 
         return resources
 
@@ -381,18 +428,8 @@ class Builder():
 
         return name
 
-    def _write_resource(self, resources_dir, resource, package, src, freeze):
-        """ Create a .qrc file for a resource and the corresponding contents.
-        """
-
-        # The main application resource does not go in a sub-directory.
-        if resource == '':
-            qrc_file = 'pyqtdeploy.qrc'
-            dst_root_dir = resources_dir
-        else:
-            qrc_file = resource + '.qrc'
-            dst_root_dir = os.path.join(resources_dir, resource)
-            self._create_directory(dst_root_dir)
+    def _write_resource_qrc(self, resources_dir, qrc_file, resource_contents):
+        """ Create a .qrc file for a resource. """
 
         f = self._create_file(resources_dir, qrc_file)
 
@@ -401,10 +438,8 @@ class Builder():
     <qresource>
 ''')
 
-        src_root_dir, src_root = os.path.split(src)
-
-        self._write_package_contents(package.contents, dst_root_dir,
-                src_root_dir, [src_root], freeze, f)
+        for content in resource_contents:
+            f.write('        <file>{0}</file>\n'.format(content))
 
         f.write('''    </qresource>
 </RCC>
@@ -412,7 +447,28 @@ class Builder():
 
         f.close()
 
-    def _write_package_contents(self, contents, dst_root_dir, src_root_dir, dir_stack, freeze, f):
+    def _write_package(self, resources_dir, resource, package, src, freeze, opt):
+        """ Write the contents of a single package and return the list of files
+        written relative to the resources directory.
+        """
+
+        # The main application resource does not go in a sub-directory.
+        if resource == '':
+            dst_root_dir = resources_dir
+        else:
+            dst_root_dir = os.path.join(resources_dir, resource)
+            self._create_directory(dst_root_dir)
+
+        src_root_dir, src_root = os.path.split(src)
+
+        resource_contents = []
+
+        self._write_package_contents(package.contents, dst_root_dir,
+                src_root_dir, [src_root], freeze, opt, resource_contents)
+
+        return resource_contents
+
+    def _write_package_contents(self, contents, dst_root_dir, src_root_dir, dir_stack, freeze, opt, resource_contents):
         """ Write the contents of a single package directory. """
 
         dir_tail = os.path.join(*dir_stack)
@@ -440,7 +496,8 @@ class Builder():
             if isinstance(content, QrcDirectory):
                 dir_stack.append(content.name)
                 self._write_package_contents(content.contents, dst_root_dir,
-                        src_root_dir, dir_stack, freeze, f)
+                        src_root_dir, dir_stack, freeze, opt,
+                        resource_contents)
                 dir_stack.pop()
             else:
                 freeze_file = True
@@ -465,44 +522,87 @@ class Builder():
 
                 file_path.append(dst_file)
 
-                f.write(
-                        '        <file>{0}</file>\n'.format(
-                                '/'.join(file_path)))
+                resource_contents.append('/'.join(file_path))
 
                 if freeze_file:
-                    self._freeze(dst_path, src_path, freeze, as_data=True)
+                    self._freeze(dst_path, src_path, freeze, opt, as_data=True)
                 else:
                     shutil.copyfile(src_path, dst_path)
 
-    @classmethod
-    def _write_main_c(cls, build_dir, app_name, extension_names):
+    def _write_main_c(self, build_dir, extension_names):
         """ Create the application specific main.c file. """
 
-        f = cls._create_file(build_dir, 'main.c')
+        project = self._project
 
-        f.write('''#include <wchar.h>
-#include <Python.h>
+        f = self._create_file(build_dir, 'main.c')
+
+        f.write('''#include <Python.h>
 
-int main(int argc, char **argv)
-{
 ''')
 
         if len(extension_names) > 0:
             inittab = 'extension_modules'
 
             f.write('#if PY_MAJOR_VERSION >= 3\n')
-            cls._write_inittab(f, extension_names, inittab, py3=True)
+            self._write_inittab(f, extension_names, inittab, py3=True)
             f.write('#else\n')
-            cls._write_inittab(f, extension_names, inittab, py3=False)
+            self._write_inittab(f, extension_names, inittab, py3=False)
             f.write('#endif\n\n')
         else:
             inittab = 'NULL'
 
-        f.write('#if PY_MAJOR_VERSION >= 3\n')
-        cls._write_main_call(f, app_name, inittab, py3=True)
-        f.write('#else\n')
-        cls._write_main_call(f, app_name, inittab, py3=False)
-        f.write('#endif\n}\n')
+        sys_path = project.sys_path
+
+        if sys_path != '':
+            f.write('static const char *path_dirs[] = {\n')
+
+            # Extract the (possibly quoted) individual directories.
+            start = -1
+            quote = ''
+
+            for i, ch in enumerate(sys_path):
+                dir_name = None
+
+                if ch == quote:
+                    dir_name = sys_path[start:i]
+                    start = -1
+                    quote = ''
+                elif ch in ('"\''):
+                    start = i + 1
+                    quote = ch
+                elif quote == '' and ch == ' ':
+                    if start != -1:
+                        dir_name = sys_path[start:i]
+                        start = -1
+                else:
+                    if start == -1:
+                        start = i
+
+                if dir_name is not None:
+                    f.write('    "{0}",\n'.format(dir_name))
+
+            if start != -1:
+                f.write('    "{0}",\n'.format(sys_path[start:]))
+
+            f.write('''    NULL
+};
+
+''')
+
+        app_name = project.application_basename()
+        path_dirs = "path_dirs" if sys_path != '' else "NULL"
+
+        f.write('''extern int pyqtdeploy_start(int argc, char **argv,
+        const char *py_main_filename, struct _inittab *extension_modules,
+        const char **path_dirs);
+''')
+
+        f.write('''
+int main(int argc, char **argv)
+{
+    return pyqtdeploy_start(argc, argv, ":/%s.pyf", %s, %s);
+}
+''' % (app_name, inittab, path_dirs))
 
         f.close()
 
@@ -520,44 +620,33 @@ int main(int argc, char **argv)
         for ext in extension_names:
             base_ext = ext.split('.')[-1]
 
-            f.write('    extern %s%s%s(void);\n' % (init_type, init_prefix,
+            f.write('extern %s%s%s(void);\n' % (init_type, init_prefix,
                     base_ext))
 
         f.write('''
-    static struct _inittab %s[] = {
+static struct _inittab %s[] = {
 ''' % inittab)
 
         for ext in extension_names:
             base_ext = ext.split('.')[-1]
 
-            f.write('        {"%s", %s%s},\n' % (ext, init_prefix, base_ext))
+            f.write('    {"%s", %s%s},\n' % (ext, init_prefix, base_ext))
 
-        f.write('''        {NULL, NULL}
-    };
+        f.write('''    {NULL, NULL}
+};
 ''')
 
-    @staticmethod
-    def _write_main_call(f, app_name, inittab, py3):
-        """ Write the Python version specific call to pyqtdeploy_main(). """
-
-        if py3:
-            name_type = 'wchar_t'
-            name_prefix = 'L'
-        else:
-            name_type = 'char'
-            name_prefix = ''
-
-        f.write('''    extern int pyqtdeploy_main(int argc, char **argv, %s *py_main,
-            struct _inittab *extension_modules);
-
-    return pyqtdeploy_main(argc, argv, %s"%s", %s);
-''' % (name_type, name_prefix, app_name, inittab))
-
-    def _freeze(self, output, py_filename, freeze, name=None, as_data=False):
+    def _freeze(self, output, py_filename, freeze, opt, name=None, as_data=False):
         """ Freeze a Python source file to a C header file or a data file. """
 
-        argv = [os.path.expandvars(self._project.python_host_interpreter),
-            freeze]
+        argv = [os.path.expandvars(self._project.python_host_interpreter)]
+
+        if opt == 2:
+            argv.append('-OO')
+        elif opt == 1:
+            argv.append('-O')
+
+        argv.append(freeze)
         
         if name is not None:
             argv.append('--name')
@@ -567,9 +656,10 @@ int main(int argc, char **argv)
         argv.append(output)
         argv.append(py_filename)
 
-        self._progress("Freezing {0}".format(py_filename))
+        self._message_handler.progress_message(
+                "Freezing {0}".format(py_filename))
 
-        self.run(argv, "Unable to freeze {0}.".format(py_filename))
+        self.run(argv, "Unable to freeze {0}".format(py_filename))
 
     def run(self, argv, error_message, in_build_dir=False):
         """ Execute a command and capture the output. """
@@ -578,37 +668,35 @@ int main(int argc, char **argv)
             saved_cwd = os.getcwd()
             build_dir = self._project.absolute_path(self._project.build_dir)
             os.chdir(build_dir)
-            self._verbose(
-                    "{0} is now the current directory.".format(build_dir))
+            self._message_handler.verbose_message(
+                    "{0} is now the current directory".format(build_dir))
         else:
             saved_cwd = None
 
-        self._verbose("Running '{0}'".format(' '.join(argv)))
+        self._message_handler.verbose_message(
+                "Running '{0}'".format(' '.join(argv)))
 
         try:
             subprocess.check_output(argv, stderr=subprocess.STDOUT,
                     universal_newlines=True)
         except subprocess.CalledProcessError as e:
-            self._error(error_message, e.output)
+            raise UserException(error_message, e.output)
+        except FileNotFoundError:
+            raise UserException("{0} does not seem to exist".format(argv[0]))
         finally:
             if saved_cwd is not None:
                 os.chdir(saved_cwd)
-                self._verbose(
-                        "{0} is now the current directory.".format(saved_cwd))
+                self._message_handler.verbose_message(
+                        "{0} is now the current directory".format(saved_cwd))
 
     @staticmethod
-    def _lib_filename(filename):
-        """ Get name of a file in the 'lib' directory that can be used by
-        QFile.
-        """
+    def _get_lib_file_name(file_name):
+        """ Get name of a file in the 'lib' sub-directory. """
 
-        lib_dir = QFileInfo(__file__).dir()
-        lib_dir.cd('lib')
-
-        return lib_dir.filePath(filename)
+        return get_embedded_dir(__file__, 'lib').absoluteFilePath(file_name)
 
     @classmethod
-    def _copy_lib_file(cls, filename, dirname=None):
+    def _copy_lib_file(cls, file_name, dir_name=None, dst_file_name=None):
         """ Copy a library file to a directory and return the full pathname of
         the copy.  If the directory wasn't specified then copy it to a
         temporary directory.
@@ -617,78 +705,40 @@ int main(int argc, char **argv)
         # Note that we use the Qt file operations to support the possibility
         # that pyqtdeploy itself has been deployed as a single executable.
 
-        # The destination filename.
-        if dirname is None:
-            dirname = tempfile.gettempdir()
+        if dir_name is None:
+            dir_name = tempfile.gettempdir()
 
-        d_filename = os.path.join(dirname, filename)
+        if dst_file_name is None:
+            dst_file_name = file_name
+            s_file_name = cls._get_lib_file_name(file_name)
+        else:
+            s_file_name = file_name
 
-        # The source filename.
-        s_filename = cls._lib_filename(filename)
+        d_file_name = os.path.join(dir_name, dst_file_name)
 
         # Make sure the destination doesn't exist.
-        QFile.remove(d_filename)
+        QFile.remove(d_file_name)
 
-        if not QFile.copy(s_filename, QDir.fromNativeSeparators(d_filename)):
-            self._error("Unable to copy file {0}.".format(filename))
+        if not QFile.copy(s_file_name, QDir.fromNativeSeparators(d_file_name)):
+            raise UserException("Unable to copy file {0}".format(file_name))
 
-        return d_filename
+        return d_file_name
 
     @staticmethod
     def _create_file(build_dir, filename):
         """ Create a text file in the build directory. """
 
-        pathname = os.path.join(build_dir, filename)
-
-        try:
-            return open(pathname, 'wt')
-        except Exception as e:
-            self._error("Unable to create file {0}.".format(pathname), str(e))
+        return create_file(os.path.join(build_dir, filename))
 
     def _create_directory(self, dir_name):
         """ Create a directory which may already exist. """
 
-        self._verbose("Creating directory {0}.".format(dir_name))
+        self._message_handler.verbose_message(
+                "Creating directory {0}".format(dir_name))
 
         try:
             os.makedirs(dir_name, exist_ok=True)
         except Exception as e:
-            self._error(
-                    "Unable to create the '{0}' directory.".format(dir_name),
+            raise UserException(
+                    "Unable to create the '{0}' directory".format(dir_name),
                     str(e))
-
-    def _progress(self, text):
-        """ Display a progress message if requested. """
-
-        if not self.quiet:
-            self.information(text)
-
-    def _verbose(self, text):
-        """ Display a verbose progress message if requested. """
-
-        if self.verbose:
-            self.information(text)
-
-    def _error(self, text, detail=''):
-        """ Handle an error.  This will raise an exception and not return. """
-
-        self.error(text)
-
-        if detail != '':
-            self.error(detail)
-
-        raise UserException(text, detail)
-
-    def information(self, text):
-        """ Handle a user information message (which will not have a trailing
-        newline).  This default implementation just sends it to stdout.
-        """
-
-        print(text)
-
-    def error(self, text):
-        """ Handle a user error message (which will not have a trailing
-        newline).  This default implementation just sends it to stderr.
-        """
-
-        print(text, file=sys.stderr)
